from .importer import main_importer, inner_importer

import argparse
from .core import JobManager
import logging as lg
import os
import shutil
from colorama import Fore
from .pogmake_execeptions import NoJobError

orig_dir = os.path.abspath(os.path.dirname(__file__))

def print_hidden_debug():
    pass


parser = argparse.ArgumentParser(description=("Registers and executes jobs"))

parser.add_argument("--print-hidden-debug", action="store_true", help=argparse.SUPPRESS)

parser.add_argument(
    "-v",
    "--verbose",
    action="store_true",
    help="Overrides logging level to DEBUG",
)

parser.add_argument(
    "--explain-depth",
    default=None,
    help="Show detailed debug information regarding a specific job",
)

parser.add_argument(
    "-e",
    "--explain",
    default=None,
    help="Show detailed debug information regarding a specific job",
)

parser.add_argument(
    "--init",
    action="store_true",
    help="Creates a default pogfile in the starting location.",
)

parser.add_argument(
    "-s",
    "--start_dir",
    default="./",
    help="This flag will cause pogmake to run in an alternate start location. As if that directory was the root directory.",
)

loglevels = {
    "DEBUG": lg.DEBUG,
    "INFO": lg.INFO,
    "WARNING": lg.WARNING,
}

parser.add_argument(
    "--log-level",
    default="INFO",
    choices=loglevels.keys(),
    help="Sets the logging level for the program's run. (gets overriden by --verbose)",
)

parser.add_argument(
    "jobs",
    nargs="*",
    default=["__ALL_JOBS__"],
    type=str,
    help="Jobs to run, defaults to all",
)

parser.add_argument(
    "-l", "--list-jobs", action="store_true", help="List all jobs available"
)

def log_setup(args):
    lg.basicConfig(level=loglevels[args.log_level])
    lg.info(f"Logging level set to: {args.log_level}")

def create_manager(args):

    args.start_dir = os.path.abspath(args.start_dir)
    dirpath = args.start_dir

    if not os.path.isdir(dirpath):
        dirpath = os.path.dirname(args.start_dir)

    lg.debug(f"Starting at path: {dirpath}.")

    gjobs = main_importer(dirpath, args)
    if len(gjobs.keys()) == 0:
        lg.error(f"{Fore.RED}: No jobs found in the root folder or any subfolders")
        raise NoJobError
    manager = JobManager(gjobs)
    
    return manager

def main():

    args = parser.parse_args()

    if args.init:
        shutil.copy(os.path.abspath(os.path.join(orig_dir, '../', 'tests/init_test/pogfile.py')), './')
        print("Check out pogfile.py or run pog --list to see available jobs.")
        return

    manager = create_manager(args)


    if args.explain:
        manager.show_detailed_info(args.explain)
        return

    if args.print_hidden_debug:
        print_debug()

    if args.list_jobs:
        manager.show_jobs()
        return

    manager.queue_jobs(args.jobs)
    manager.run_jobs()


def gmain():
    """
    This is the gui version of the pogmake frontend

    Not available yet.
    """
    pass



# pogfile shared imports
import subprocess
import sys
import argparse
import os
import shutil

# /pogfile shared imports

import inspect
import importlib
import logging as lg

from colorama import Fore, Style, init

init(autoreset=True)

gargs = None
g_jobs = {}

def get_gjobs():
    global g_jobs
    return g_jobs


def get_gargs():
    global gargs
    return gargs


class JobInfo:
    def __init__(self, func, desc, deps=None, default=False, name="Unknown", **kwargs):
        self.func = func
        self.docs = desc
        self.name = name
        self.deps = [*deps]
        # print("deps : ",self.deps)
        self.default = default
        self.extra = kwargs

    def explain_deps_nested(self, jobdict, _nest_count=None, _superiors=None):

        nest_count = 0 if _nest_count is None else _nest_count
        superiors = [] if _superiors is None else _superiors
        dash = "- "
        nest_count += 1
        for dep in self.deps:
            if dep not in jobdict:
                print(f"{' '*nest_count*2}{dash}{Fore.RED}{dep} (undefined dependency)")
            else:
                depinfo = jobdict[dep]
                descstr = "" if depinfo.docs == None else f": {Fore.CYAN}{depinfo.docs}"
                print(f"{' '*nest_count*2}{dash}{Fore.GREEN}{dep}{descstr}")
                if dep not in superiors:
                    superiors.append(dep)
                    jobdict[dep].explain_deps_nested(jobdict, nest_count, superiors)

    def get_deplist(self, deplist=None):
        global g_jobs

        if deplist is None:
            deplist = []

        for dep in self.deps:
            if not dep in g_jobs:
                print(
                    f"{Fore.RED}{dep} is listed as a dep of {self.name}, but is not a valid target"
                )
                continue

            if dep == self.name:
                print(f"warning: {self.name} has itself listed as a dependency")
                continue
            if dep not in deplist:
                deplist.append(dep)
                deplist += g_jobs[dep].get_deplist(deplist)

        odeplist = []

        for dep in deplist:
            if dep not in odeplist:
                odeplist.append(dep)
        return odeplist


def job(*deps, desc=None, default=True):
    def wrap(f):
        global g_jobs
        g_jobs[f.__name__] = JobInfo(
            f, desc, deps, default, f.__name__, where_defined=f.__code__.co_filename
        )

        def wrapped_f(*deps):
            f(*deps)

        return wrapped_f

    return wrap


class JobManager:
    def __init__(self, jobs):
        self.jobs = jobs
        self.defaults = []

        for name, info in self.jobs.items():
            if info.default:
                self.defaults.append(name)

        self.dispatched_jobs = []
        self.queued_jobs = []
        self.n_jobs = 0
        self.completed_jobs = []

    def queue_jobs(self, joblist):
        queued_jobs = joblist

        if "__ALL_JOBS__" in joblist:
            queued_jobs = self.defaults

        for jobname in queued_jobs:
            jobinfo = self.jobs[jobname]
            deplist = jobinfo.get_deplist()
            for dep in deplist:
                if dep not in queued_jobs:
                    queued_jobs.append(dep)

        self.queued_jobs = queued_jobs
        self.queued_count = len(queued_jobs)

        print("======================================================================")
        print("Queueing the following jobs:")
        maxshow = 10
        _maxshow = maxshow
        for job in self.queued_jobs:
            print(f"    {Fore.CYAN}{job}")
            maxshow -= 1
            if maxshow == 0:
                print(
                    f"    ... and {Fore.YELLOW}{self.queued_count - _maxshow}{Fore.RESET} more"
                )
                break
        print(f"  TOTAL: {Fore.YELLOW}{self.queued_count}")
        print("======================================================================")
        return queued_jobs

    def show_detailed_info(self, jobname):
        if jobname not in self.jobs:
            print(f"{Fore.RED}{jobname} is not a registered job")
            return
        job = self.jobs[jobname]
        print(f"Job: {Fore.CYAN}{jobname}")
        print(f"File: {Fore.CYAN}{job.extra['where_defined']}")
        print(f"Depends: ")
        job.explain_deps_nested(self.jobs)

    def show_jobs(self):
        word = "are"
        s = "s"
        default_jobs = []
        nondefault_jobs = []
        for name, job in self.jobs.items():
            if job.default:
                default_jobs.append(job)
            else:
                nondefault_jobs.append(job)

        if len(self.jobs) == 1:
            word = "is"
            s = ""

        def print_job(job):
            name = job.name
            asterix = "* "
            if not job.default:
                asterix = "  "
            else:
                asterix = "* "

            if job.docs is not None:
                print(f"   {asterix}{Fore.MAGENTA + name} - {Fore.GREEN + job.docs}")
            else:
                print(
                    f"   {asterix}{Fore.MAGENTA + name} - {Fore.YELLOW + 'no description available'}"
                )

        print(
            "\n======================================================================"
        )
        print(
            f"There {word} {Fore.YELLOW}{len(self.jobs)}{Style.RESET_ALL} job{s} registered: (* means part of the default)"
        )

        for job in default_jobs:
            print_job(job)

        for job in nondefault_jobs:
            print_job(job)
        print("======================================================================")

    def run_job(self, name):
        if name not in self.jobs:
            print(
                f"{Fore.RED}'{name}' is a listed job but we're skipping it because it is undefined..."
            )
            return

        job = self._get_job(name)

        if name not in self.completed_jobs:
            for dep in job.deps:
                if name == dep:
                    continue
                if dep not in self.completed_jobs:
                    self.run_job(dep)
            print(
                Fore.MAGENTA
                + f"=========== {Fore.GREEN}Running Job: "
                + Fore.CYAN
                + name
                + Fore.MAGENTA
                + f" [{Fore.GREEN}{len(self.completed_jobs) + 1}{Fore.MAGENTA}/{self.queued_count}] ======== "
            )
            try:
                job.func()
                self.completed_jobs.append(name)
            except:
                print(f"{Fore.RED}Job {name} failed")
                raise

    def run_jobs(self):
        for job in self.queued_jobs:
            self.run_job(job)

    def _get_job(self, jobname):
        job = self.jobs[jobname]
        return job

def get_manager():
    return JobManager(g_jobs)




from .core import JobManager, job, get_manager, get_gjobs
from .app import main, parser, create_manager
from .pogmake_execeptions import NoJobError
from .jobenv import JobEnv

if __name__ == "__main__":
    main()
